from __future__ import unicode_literals

import logging, json, urllib, urllib2
import tornado.web
from spotipy import Spotify

# import logger
logger = logging.getLogger(__name__)
        
class RequestHandler(tornado.web.RequestHandler):

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")

    def initialize(self, core, config):
        self.core = core
        self.config = config
    
    def get(self, slug=None):

        if( slug == 'refresh_spotify_token' ):
            self.write( self.refresh_spotify_token() )
            return

        else:
            self.write('Invalid request')
            return


    ##
    # Get a new spotify authentication token
    #
    # Uses the Client Credentials Flow, so is invisible to the user. We need this token for
    # any backend spotify requests (we don't tap in to Mopidy-Spotify, yet). Also used for
    # passing token to frontend for javascript requests without use of the Authorization Code Flow.
    ##
    def refresh_spotify_token( self ):
    
        url = 'https://accounts.spotify.com/api/token'
        authorization = 'YTg3ZmI0ZGJlZDMwNDc1YjhjZWMzODUyM2RmZjUzZTI6ZDdjODlkMDc1M2VmNDA2OGJiYTE2NzhjNmNmMjZlZDY='

        headers = {'Authorization' : 'Basic ' + authorization}
        data = {'grant_type': 'client_credentials'}
        data_encoded = urllib.urlencode( data )
        req = urllib2.Request(url, data_encoded, headers)

        try:
            response = urllib2.urlopen(req, timeout=30).read()
            response = json.loads(response)
            return response
        except urllib2.HTTPError as e:
            return e
        



import tornado.ioloop, tornado.web, tornado.websocket, tornado.template
import logging, uuid, subprocess, pykka
from datetime import datetime
from tornado.escape import json_encode, json_decode

logger = logging.getLogger(__name__)

# container for all current pusher connections
connections = {}
frontend = {}
  
  
##
# Send a message to an individual connection
#
# @param recipient_connection_ids = array
# @param action = string (action method of this message)
# @param request_id = string (used for callbacks)
# @param data = array (any data required to include in our message)
##
def send_message( recipient_connection_id, action, request_id, data ):          
    message = {
        'action': action,
        'request_id': request_id,
        'data': data
    }
    connections[recipient_connection_id]['connection'].write_message( json_encode(message) )
        
##
# Broadcast a message to all recipients
#
# @param action = string
# @param data = array (the body of our message to send)
##
def broadcast( type, data ):    
    for connection in connections.itervalues():
        message = {
            'action': 'broadcast',
            'type': type,
            'data': data
        }
        connection['connection'].write_message( json_encode(message) )
        
        
# digest a protocol header into it's id/name parts
def digest_protocol( protocol ):
    
    # if we're a string, split into list
    # this handles the different ways we get this passed (select_subprotocols gives string, headers.get gives list)
    if isinstance(protocol, basestring):
    
        # make sure we strip any spaces (IE gives "element,element", proper browsers give "element, element")
        protocol = [i.strip() for i in protocol.split(',')]
    
    # if we've been given a valid array
    try:
		clientid = protocol[0]
		connectionid = protocol[1]
		username = protocol[2]
		generated = False
      
    # invalid, so just create a default connection, and auto-generate an ID
    except:
		clientid = str(uuid.uuid4().hex)
		connectionid = str(uuid.uuid4().hex)
		username = str(uuid.uuid4().hex)
		generated = True
    
    # construct our protocol object, and return
    return {"clientid": clientid, "connectionid": connectionid, "username": username, "generated": generated}

    
##
# Websocket server
#
# This is the actual websocket thread that accepts, digests and emits messages.
# TODO: Figure out how to merge this into the main Mopidy websocket to avoid needing two websocket servers
##    
class PusherWebsocketHandler(tornado.websocket.WebSocketHandler):
    
    def initialize(self, frontend):
        self.frontend = frontend

    def check_origin(self, origin):
        return True
  
    # when a new connection is opened
    def open(self):
    
        # decode our connection protocol value (which is a payload of id/name from javascript)
        protocolElements = digest_protocol(self.request.headers.get('Sec-Websocket-Protocol', []))

        connectionid = protocolElements['connectionid']
        clientid = protocolElements['clientid']
        self.connectionid = connectionid
        username = protocolElements['username']
        created = datetime.strftime(datetime.now(), '%Y-%m-%d %H:%M:%S')

        # construct our client object, and add to our list of connections
        client = {
            'clientid': clientid,
            'connectionid': connectionid,
            'username': username,
            'ip': self.request.remote_ip,
            'created': created
        }
        connections[connectionid] = {
            'client': client,
            'connection': self
        }

        logger.info( 'Pusher connection established: '+ connectionid +' ('+ clientid +'/'+ username +')' )

        # broadcast to all connections that a new user has connected
        broadcast( 'new_connection', client )
  
    def select_subprotocol(self, subprotocols):
        # select one of our subprotocol elements and return it. This confirms the connection has been accepted.
        protocols = digest_protocol( subprotocols )

        # if we've auto-generated some ids, the provided subprotocols was a string, so just return it right back
        # this allows a connection to be completed
        if protocols['generated']:
            return subprotocols[0]
            
        # otherwise, just return one of the supplied subprotocols
        else:
            return protocols['clientid']
  
    # server received a message
    def on_message(self, message):
        messageJson = json_decode(message)

        # construct the origin client info
        messageJson['origin'] = {
            'connectionid' : self.connectionid,
            'clientid': connections[self.connectionid]['client']['clientid'],
            'ip': self.request.remote_ip,
            'username': connections[self.connectionid]['client']['username']
        }
        
        logger.debug('Pusher message received: '+message)
    
        # broadcast message to other connections (except for self)
        if messageJson['action'] == 'broadcast':

            # respond to request with status update
            send_message(
                self.connectionid, 
                'response', 
                messageJson['request_id'], 
                { 'status': 'Ok' }
            )

            for connection in connections.itervalues():
                if connection['client']['connectionid'] != self.connectionid:
                    connection['connection'].write_message(messageJson)
    
        # send authroization details
        elif messageJson['action'] == 'send_authorization':

            # make sure we actually have a connection matching the provided connectionid
            if messageJson['recipient_connectionid'] in connections:

                # send payload to recipient
                authorization_message = {
                    'type': 'broadcast',
                    'action': 'received_authorization',
                    'authorization': messageJson['authorization'],
                    'me': messageJson['me'],
                    'origin': messageJson['origin']
                }
                connections[messageJson['recipient_connectionid']]['connection'].write_message(authorization_message)

                # respond to request with status update
                send_message(
                    self.connectionid, 
                    'response', 
                    messageJson['request_id'], 
                    { 'status': 'Ok' }
                )
            else:
                # respond to request with status update
                send_message(
                    self.connectionid, 
                    'response', 
                    messageJson['request_id'], 
                    { 'error': 'Could not send to that connection, does not exist' }
                )

        # fetch our pusher connections
        elif messageJson['action'] == 'get_config':                
            send_message(
                self.connectionid, 
                'response', 
                messageJson['request_id'], 
                { 'config': self.frontend.get_config() }
            )

        # fetch our pusher connections
        elif messageJson['action'] == 'get_connections':
        
            connectionsDetailsList = []
            for connection in connections.itervalues():
                connectionsDetailsList.append(connection['client'])
                
            send_message(
                self.connectionid, 
                'response', 
                messageJson['request_id'], 
                { 'connections': connectionsDetailsList }
            )
        
        # change connection's client username
        elif messageJson['action'] == 'set_username':
            
            # username is the only value we allow clients to change
            connections[messageJson['origin']['connectionid']]['client']['username'] = messageJson['username']
            
            # respond to request
            send_message(
                self.connectionid, 
                'response',
                messageJson['request_id'],
                { 'username': messageJson['username'] }
            )
            
            # notify all clients of this change
            broadcast( 'connection_updated', { 'connection': connections[messageJson['origin']['connectionid']]['client'] })
    
        # start radio
        elif messageJson['action'] == 'start_radio':
        
            # pull out just the radio data (we don't want all the request_id guff)
            radio = {
                'enabled': 1,
                'seed_artists': messageJson['seed_artists'],
                'seed_genres': messageJson['seed_genres'],
                'seed_tracks': messageJson['seed_tracks']
            }
            radio = self.frontend.start_radio( radio )
            send_message( 
                self.connectionid, 
                'response',
                messageJson['request_id'], 
                { 'radio': radio }
            )
    
        # stop radio
        elif messageJson['action'] == 'stop_radio':
            radio = self.frontend.stop_radio()
            send_message( 
                self.connectionid, 
                'response', 
                messageJson['request_id'], 
                { 'radio': self.frontend.radio }
            )
        
        # fetch our current radio state
        elif messageJson['action'] == 'get_radio':
            send_message( 
                self.connectionid, 
                'response',
                messageJson['request_id'],
                { 'radio': self.frontend.radio }
            )
    
        # get system version and check for upgrade
        elif messageJson['action'] == 'get_version':
            version = self.frontend.get_version()
            send_message( 
                self.connectionid, 
                'response',
                messageJson['request_id'], 
                { 'version': version } 
            )
    
        # perform upgrade
        elif messageJson['action'] == 'upgrade':
            version = self.frontend.get_version()
            upgrade_successful = self.frontend.perform_upgrade()
            send_message( 
                self.connectionid, 
                'response',
                messageJson['request_id'], 
                { 'upgrade_successful': upgrade_successful, 'version': version } 
            )
    
        # restart mopidy
        elif messageJson['action'] == 'restart':
            self.frontend.restart()            
        
        # not an action we recognise!
        else:
            send_message( 
                self.connectionid, 
                'response', 
                messageJson['request_id'], 
                { 'error': 'Unhandled action' } 
            )
                        
        logger.debug( 'Pusher: Unhandled message received from '+ self.connectionid )
  
    # connection closed
    def on_close(self):
        if self.connectionid in connections:
            
            clientRemoved = connections[self.connectionid]['client']
            logger.debug( 'Spotmop Pusher connection to '+ self.connectionid +' closed' )
            
            # now actually remove it
            try:
                del connections[self.connectionid]
            except:
                logger.info( 'Failed to close connection to '+ self.connectionid )                
            
            broadcast( 'client_disconnected', clientRemoved )
        
        
        
  


from __future__ import unicode_literals

import logging, os, json
import tornado.web
import tornado.websocket
from mopidy import config, ext
from frontend import IrisFrontend
from http import RequestHandler

logger = logging.getLogger(__name__)
__version__ = '2.11.5'

##
# Core extension class
#
# Loads config and gets the party started. Initiates any additional frontends, etc.
##
class Extension( ext.Extension ):

    dist_name = 'Mopidy-Iris'
    ext_name = 'iris'
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = config.ConfigSchema(self.ext_name)
        schema['enabled'] = config.Boolean()
        schema['pusherport'] = config.String()
        schema['country'] = config.String()
        schema['locale'] = config.String()
        return schema

    def setup(self, registry):
        
        # Add web extension
        registry.add('http:app', {
            'name': self.ext_name,
            'factory': factory
        })
        
        # add our frontend
        registry.add('frontend', IrisFrontend)
        
def factory(config, core):

    path = os.path.join( os.path.dirname(__file__), 'static')
	
    return [
        (r"/images/(.*)", tornado.web.StaticFileHandler, {
            "path": config['local-images']['image_dir']
        }),
        (r'/http/([^/]*)', RequestHandler, {
                'core': core,
                'config': config
            }),
        (r'/(.*)', tornado.web.StaticFileHandler, {
				"path": path,
				"default_filename": "index.html"
			}),
    ]



from __future__ import unicode_literals

import logging, json, pykka, pylast, pusher, urllib, urllib2, os, sys, mopidy_iris, subprocess
import tornado.web
import tornado.websocket
import tornado.ioloop
from mopidy import config, ext
from mopidy.core import CoreListener
from pkg_resources import parse_version
from spotipy import Spotify

# import logger
logger = logging.getLogger(__name__)
    
###
# Spotmop supporting frontend
#
# This provides a wrapping thread for the Pusher websocket, as well as the radio infrastructure
##
class IrisFrontend(pykka.ThreadingActor, CoreListener):

    def __init__(self, config, core):
        global spotmop
        super(IrisFrontend, self).__init__()
        self.config = config
        self.core = core
        self.version = mopidy_iris.__version__
        self.is_root = ( os.geteuid() == 0 )
        self.spotify_token = False
        self.radio = {
            "enabled": 0,
            "seed_artists": [],
            "seed_genres": [],
            "seed_tracks": []
        }

    def on_start(self):
        
        logger.info('Starting Iris '+self.version)
        
        # try and start a pusher server
        port = str(self.config['iris']['pusherport'])
        try:
            self.pusher = tornado.web.Application([( '/pusher', pusher.PusherWebsocketHandler, { 'frontend': self } )])
            self.pusher.listen(port)
            logger.info('Pusher server running at [0.0.0.0]:'+port)
            
        except( pylast.NetworkError, pylast.MalformedResponseError, pylast.WSError ) as e:
            logger.error('Error starting Pusher: %s', e)
            self.stop()
        
        # get a fresh spotify authentication token and store for future use
        # self.refresh_spotify_token()


    ##
    # Get a new spotify authentication token for server-side use
    #
    # Uses the Client Credentials Flow, so is invisible to the user. We need this token for
    # any backend spotify requests (we don't tap in to Mopidy-Spotify, yet). Also used for
    # passing token to frontend for javascript requests without use of the Authorization Code Flow.
    ##
    def refresh_spotify_token( self ):
    
        url = 'https://accounts.spotify.com/api/token'
        authorization = 'YTg3ZmI0ZGJlZDMwNDc1YjhjZWMzODUyM2RmZjUzZTI6ZDdjODlkMDc1M2VmNDA2OGJiYTE2NzhjNmNmMjZlZDY='

        headers = {'Authorization' : 'Basic ' + authorization}
        data = {'grant_type': 'client_credentials'}
        data_encoded = urllib.urlencode( data )
        req = urllib2.Request(url, data_encoded, headers)

        try:
            response = urllib2.urlopen(req, timeout=30).read()
            response_dict = json.loads(response)
            self.spotify_token = response_dict
            return response_dict
        except urllib2.HTTPError as e:
            return e

    
    ##
    # Listen for core events, and update our frontend as required
    ##
    def track_playback_ended( self, tl_track, time_position ):
        self.check_for_radio_update()        
        
        
    ##
    # See if we need to perform updates to our radio
    #
    # We see if we've got one or two tracks left, if so, go get some more
    ##
    def check_for_radio_update( self ):
        try:
            tracklistLength = self.core.tracklist.length.get()        
            if( tracklistLength <= 5 and self.radio['enabled'] == 1 ):
                self.load_more_tracks()
                
        except RuntimeError:
            pusher.broadcast('error', {'source': 'check_for_radio_update', 'message': 'Could not fetch tracklist length'})
            logger.warning('IrisFrontend: Could not fetch tracklist length')
            pass


    ##
    # Load some more radio tracks
    #
    # We need to build a Spotify authentication token first, and then fetch recommendations
    ##
    def load_more_tracks( self ):
        
        # this is crude, but it means we don't need to handle expired tokens
        # TODO: address this when it's clear what Jodal and the team want to do with Pyspotify
        self.refresh_spotify_token()
        
        try:
            token = self.spotify_token
            token = token['access_token']
        except:
            logger.error('IrisFrontend: access_token missing or invalid')
            pusher.broadcast('error', {'source': 'load_more_tracks', 'message': 'access_token missing or invalid'})
            
        try:
            spotify = Spotify( auth = token )
            response = spotify.recommendations(seed_artists = self.radio['seed_artists'], seed_genres = self.radio['seed_genres'], seed_tracks = self.radio['seed_tracks'], limit = 5)
            
            uris = []
            for track in response['tracks']:
                uris.append( track['uri'] )
            
            self.core.tracklist.add( uris = uris )
        except:
            pusher.broadcast('error', {'source': 'load_more_tracks', 'message': 'Failed to fetch Spotify recommendations'})
            logger.error('IrisFrontend: Failed to fetch Spotify recommendations')
            
    
    ##
    # Start radio
    #
    # Take the provided radio details, and start a new radio process
    ##
    def start_radio( self, new_state ):
        
        # set our new radio state
        self.radio = new_state
        self.radio['enabled'] = 1;
        
        # clear all tracks
        self.core.tracklist.clear()
        
        # explicitly set consume, to ensure we don't end up with a huge tracklist (and it's how a radio should 'feel')
        self.core.tracklist.set_consume( True )
        
        # load me some tracks, and start playing!
        self.load_more_tracks()
        self.core.playback.play()
        
        # notify clients
        pusher.broadcast('radio', { 'radio': self.radio })
        
        # return new radio state to initial call
        return self.radio
        
    ##
    # Stop radio
    ##
    def stop_radio( self ):
        
        # reset radio
        self.radio = {
            "enabled": 0,
            "seed_artists": [],
            "seed_genres": [],
            "seed_tracks": []
        }
        
        # stop track playback
        self.core.playback.stop()

        # notify clients
        pusher.broadcast( 'radio', { 'radio': self.radio })
        
        # return new radio state to initial call
        return self.radio
        
   
    # get our spotify token
    def get_spotify_token( self ):
        return self.spotify_token
        
   
    # get our config values
    def get_config( self ):
        all_config = self.config
        config = {
            "spotify_username": all_config['spotify']['username'],
            "country": all_config['iris']['country'],
            "locale": all_config['iris']['locale']
        }
        return config
        
        
    ##
    # Get Spotmop version, and check for updates
    #
    # We compare our version with the latest available on PyPi
    ##
    def get_version( self ):
        
        url = 'https://pypi.python.org/pypi/Mopidy-Iris/json'
        req = urllib2.Request(url)
        
        try:
            response = urllib2.urlopen(req, timeout=30).read()
            response = json.loads(response)
            latest_version = response['info']['version']
            
            # compare our versions, and convert result to boolean
            upgrade_available = cmp( parse_version( latest_version ), parse_version( self.version ) )
            upgrade_available = ( upgrade_available == 1 )

        except urllib2.HTTPError as e:
            latest_version = '0.0.0'
            upgrade_available = False
        
        # prepare our response
        data = {
            'current': self.version,
            'latest': latest_version,
            'is_root': self.is_root,
            'upgrade_available': upgrade_available
        }
        return data
        
        
    ##
    # Upgrade Spotmop module
    #
    # Upgrade myself to the latest version available on PyPi
    ##
    def perform_upgrade( self ):
        try:
            subprocess.check_call(["pip", "install", "--upgrade", "Mopidy-Iris"])
            return True
        except subprocess.CalledProcessError:
            return False
        
    ##
    # Restart Mopidy
    #
    # This is untested and may require installation of an upstart script to properly restart
    ##
    def restart( self ):
        os.execl(sys.executable, *([sys.executable]+sys.argv))
        